import typer
import os
import shutil
import locale
from pathlib import Path

APP_NAME = "ntuee-mailer"
APP_ROOT = Path(os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__))))
APP_DIR = Path(typer.get_app_dir(APP_NAME))
# make APP_DIR if it doesn't exist
APP_DIR.mkdir(parents=True, exist_ok=True)
CONFIG_PATH = Path(APP_DIR) / "config.ini"
if not CONFIG_PATH.is_file():
    shutil.copy(APP_ROOT / "config-default.ini", CONFIG_PATH)

ENCODING = locale.getpreferredencoding()



##########################################################################
# File:         AutoMailer.py                                            #
# Purpose:      Automatically send batch of mails                        #
# Last changed: 2015/06/21                                               #
# Author:       zhuang-jia-xu                                            #
# Edited:                                                                #
# Copyleft:     (ɔ)NTUEE                                                 #
##########################################################################
from rich import print
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeRemainingColumn,
    BarColumn,
)
from rich.prompt import Prompt
from cerberus import Validator

import time
import os
import re
import logging
from typing import List
from configparser import ConfigParser
from pathlib import Path
import smtplib
from email.mime.multipart import MIMEMultipart
import poplib
from email.parser import Parser as EmailParser

from .utils import *
from .globals import *
from .Letter import Letter

__all__ = ["AutoMailer"]

auto_mailer_config_schema = {
    "account": {
        "type": "dict",
        "schema": {
            "name": {"type": "string", "required": True},
            "userid": {"type": "string"},
        },
    },
    "smtp": {
        "require_all": True,
        "type": "dict",
        "schema": {
            "host": {"type": "string"},
            "port": {"type": "integer", "coerce": int},
            "timeout": {"type": "integer", "coerce": int},
        },
    },
    "pop3": {
        "require_all": True,
        "type": "dict",
        "schema": {
            "host": {"type": "string"},
            "port": {"type": "integer", "coerce": int},
            "timeout": {"type": "integer", "coerce": int},
        },
    },
}
v = Validator(auto_mailer_config_schema)

email_re = re.compile("[a-z0-9-_\.]+@[a-z0-9-\.]+\.[a-z\.]{2,5}")


class AutoMailer:
    verbose: bool = True
    SMTPserver: smtplib.SMTP_SSL = None
    config: dict = None
    total_count: int = 0
    success_count: int = 0
    email_addrs: List[str] = []
    userid: str = None
    password: str = None

    def __init__(self, config: dict = None, quiet: bool = False) -> None:
        self.config = config
        self.verbose = not quiet
        self.SMTPserver = self.__createSMTPServer()

    def login(self) -> None:
        """login to SMTP server"""
        for i in range(3):
            try:
                self.SMTPserver.login(*self.__get_login_info())
            except KeyboardInterrupt:
                exit(1)
            except:
                logging.info(f"Login failed {i+1} times")
                richError(
                    "\nLogin failed, please try again", prefix="", terminate=False
                )
                time.sleep(1)
                continue
            richSuccess("Login success")
            return

        richError("Too many failed login attempts", prefix="")

    def __get_login_info(self) -> dict:
        """get login info"""
        print("\n[blue]\[User login]")

        if "userid" in self.config["account"]:
            userid = Prompt.ask(
                "[blue]Username[/blue] saved user id:",
                default=self.config["account"]["userid"],
            )
        else:
            userid = Prompt.ask(
                "[blue]Username[/blue] (e.g. b09901000)",
            )
            userid = userid.strip()

            save_id = Confirm.ask("[blue]remember this userid?[/blue]", default=True)
            if save_id:
                new_config = self.config.copy()
                new_config["account"] = self.config["account"].copy()
                new_config["account"]["userid"] = userid
                self.save_config(new_config)

        password = Prompt.ask("[blue]password", password=True)

        print("\n")

        self.userid = userid
        self.password = password
        return userid, password

    def send_emails(self, letter: Letter, *, test_mode: bool = False, dry: bool = False) -> None:
        """send emails"""
        if self.verbose:
            print("-" * 50)
            print(Path(letter.paths["content"]).read_text())
            print("-" * 50)
            if not countdownConfirm(
                "Are you sure to send emails with the content above?",
                default=False,
                countdown=3,
            ):
                logging.info("User cancelled on checking content")
                richError("Canceled", prefix="")
        
        if not Confirm.ask(
            f"""
You are about to send email{'s' if len(letter) > 1 else ''} 
with your name set to [blue]'{self.config['account']['name']}'[/blue]
to [blue]{len(letter)}[/blue] recipients? (please use test mode before you send emails)\n
Do you want to continue?""",
            default=False,
        ):
            logging.info("User cancelled on sending emails")
            richError("Canceled", prefix="")

        if self.SMTPserver is None:
            logging.info("SMTP server is not connected, please connect first")
            richError("SMTP server is not connected, please connect first")

        self.email_addrs += letter.email_addrs

        letter.set_from_addr(complete_school_email(self.userid))

        progress = Progress(
            TextColumn("[bold blue]{task.description}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.completed} of {task.total:.0f}",
            "•",
            TimeRemainingColumn(),
        )
        
        with progress:
            logging.info(f"Sending {len(letter)} emails")
            for email in progress.track(letter, description="Sending emails..."):
                self.__server_rest(progress)
                
                if not dry:
                    success = self.send_email(email, test_mode=test_mode)
                else:
                    success = True

                if success:
                    if self.verbose:
                        progress.print(
                            f"[green]successfully sent email to {(complete_school_email(self.userid)+' (yourself)') if test_mode else email['To']}"
                        )
                else:
                    progress.print(f"[red]failed to send email to {(complete_school_email(self.userid)+' (yourself)') if test_mode else email['To']}")

            if dry:
                print("[red]This is a dry run, no emails were actually sent")


    def send_email(self, email: MIMEMultipart, *, test_mode: bool=False) -> None:
        """send email"""
        if self.SMTPserver is None:
            richError("SMTP server is not connected, please connect first")

        self.total_count += 1

        if test_mode:
            toaddrs = complete_school_email(self.userid)
            ccaddrs = ""
            bccaddrs = ""
        else:
            toaddrs = email["To"].split(",")
            ccaddrs = email["Cc"].split(",") if email["Cc"] is not None else []
            bccaddrs = email["Bcc"].split(",") if email["Bcc"] is not None else []

        try:
            self.SMTPserver.sendmail(
                email["From"],
                toaddrs + ccaddrs + bccaddrs,
                email.as_string(),
            )
        except Exception as e:
            logging.error(e)
            logging.error(f"Failed to send email to {email['To']}")
            return False

        logging.info(f"Sent email {toaddrs}")

        self.success_count += 1
        return True

    def check_bounce_backs(self) -> None:
        """show help message if emails are bounced back, this usually happens when trying to email a wrong school email address"""
        if self.total_count == 0:
            print("No emails were sent, nothing to check")
            return

        logging.info("Checking bounce backs")
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
        )
        with progress:
            progress.add_task(description="checking for bounce-backs...", total=None)
            time.sleep(5)  # wait for bounce back

            try:
                # connect to pop3 server
                pop3 = poplib.POP3_SSL(
                    host=self.config["pop3"]["host"],
                    port=self.config["pop3"]["port"],
                    timeout=self.config["pop3"]["timeout"],
                )
                pop3.user(self.userid)
                pop3.pass_(self.password)
            except Exception as e:
                logging.error(e)
                logging.error("Failed to connect to pop3 server")
                progress.print("[red]Failed to connect to pop3 server")
                return 0

            progress.print("Connected to POP3 server")
            logging.info("Connected to POP3 server")
            # retrieve last n emails
            _, mails, _ = pop3.list()
            emails = [
                pop3.retr(i)[1]
                for i in range(len(mails), len(mails) - len(self.email_addrs), -1)
            ]
            pop3.quit()

            email_contents = []
            # Concat message pieces:
            for msg in emails:
                # some chinese character may not be able to parse,
                # however, we only care about the bounce back notifications,
                # which are always in English
                try:
                    email_contents.append(b"\r\n".join(msg).decode(ENCODING))
                except:
                    continue

            # Parse message into an email object:
            email_contents = [
                EmailParser().parsestr(content, headersonly=True)
                for content in email_contents
            ]

            bounced_list = []

            for content in email_contents:
                if not re.match(
                    "(Delivery Status Notification)|(Undelivered Mail Returned to Sender)",
                    content["subject"],
                ):
                    continue

                # match for email addresses
                addr = email_re.search(content.get_payload()).group()
                bounced_list.append(addr)

            bounced_list = list(filter(lambda x: x in self.email_addrs, bounced_list))

            if len(bounced_list) > 0:
                progress.print(
                    "[red]Emails sent to these addresses are bounced back (failed):"
                )
                for address in bounced_list:
                    progress.print(f"\t{address},")
                progress.print("[red]Please check these emails.")
            else:
                progress.print(
                    "[green]No bounce-backs found, all emails are delivered successfully",
                )

        self.success_count -= len(bounced_list)

    def __server_rest(self, progress):
        """for bypassing email server limitation"""
        def rest(seconds):
            progress.print(f"[blue]resting for {seconds} seconds...")
            time.sleep(seconds)

        if self.total_count % 260 == 0 and self.total_count > 0:
            rest(50)
        elif self.total_count % 130 == 0 and self.total_count > 0:
            rest(30)
        elif self.total_count % 10 == 0 and self.total_count > 0:
            rest(10)

    def __createSMTPServer(self) -> smtplib.SMTP_SSL:
        """create SMTP server"""

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Connecting to SMTP Server...", total=None)

            time.sleep(1)

            try:
                server = smtplib.SMTP_SSL(
                    host=self.config["smtp"]["host"],
                    port=self.config["smtp"]["port"],
                    timeout=self.config["smtp"]["timeout"],
                )
            except Exception as e:
                logging.critical(e)
                logging.critical("Failed to connect to SMTP server")
                progress.print("[red]Failed to connect to SMTP server")
                exit(1)

            if server.has_extn("STARTTLS"):
                server.starttls()

            server.ehlo_or_helo_if_needed()

        logging.info("Connected to SMTP server")
        richSuccess("SMTP server connected")

        return server

    @classmethod
    def load_mailer_config(cls, config_path: str) -> dict:
        """load auto mailer configuration from config.ini"""
        config_path = Path(config_path)
        automailer_config = ConfigParser()
        if not os.path.exists(config_path):
            richError(f"{config_path} not found")

        automailer_config.read({config_path}, encoding=ENCODING)

        sections = automailer_config.sections()

        config = {}

        for section in sections:
            options = automailer_config.options(section)
            temp_dict = {}
            for option in options:
                temp_dict[option] = automailer_config.get(section, option)

            config[section] = temp_dict

        if cls.validate_config(config, verbose=True):
            config = v.document.copy()
            if config["account"]["name"] == "":
                config["account"]["name"] = Prompt.ask(
                    'Your name is currently set to [blue]""[/blue], Please enter your name',
                )
            return config
        else:
            logging.critical(
                f"mailer config validation failed, please check {config_path}"
            )
            logging.critical(config)
            richError(
                f"mailer config validation failed, please check {config_path}",
            )

    @classmethod
    def validate_config(cls, config: dict, verbose=False) -> bool:
        """validate config"""
        if v.validate(config):
            return True
        else:
            if verbose:
                parse_validation_error(v._errors)
            return False

    @classmethod
    def save_config(cls, config: dict):
        if not cls.validate_config(config):
            return False

        new_config_parser = ConfigParser()

        for section, vals in config.items():
            new_config_parser[section] = vals

        with open(CONFIG_PATH, "w") as f:
            new_config_parser.write(f)

        return True


if __name__ == "__main__":
    auto_mailer_config = AutoMailer.load_mailer_config("config.ini")
    auto_mailer = AutoMailer(auto_mailer_config)
    emails = Letter(
        "letters/test",
        auto_mailer_config["account"]["name"],
        complete_school_email(auto_mailer_config["account"]["userid"]),
    )
    auto_mailer.connect()
    auto_mailer.send_emails(emails)
    auto_mailer.check_bounce_backs()



##########################################################################
# File:         main.py                                                  #
# Purpose:      Automatically send batch of mails                        #
# Last changed: 2022/07/16                                               #
# Author:       zhuang-jia-xu                                            #
# Edited:                                                                #
# Copyleft:     (ɔ)NTUEE                                                 #
##########################################################################
import typer
from rich import print
from rich.prompt import Confirm, Prompt

import os
import logging
import shutil
import time
from pathlib import Path
from typing import Optional
from configparser import ConfigParser

from .utils import *
from .AutoMailer import AutoMailer
from .Letter import Letter
from .globals import *

app = typer.Typer()


@app.command()
def send(
    letter_path: Optional[Path] = typer.Argument(
        None, help="Path to letter", exists=True, file_okay=False
    ),
    test_mode: bool = typer.Option(
        False, "--test", "-t", help="Test mode: send mail to yourself"
    ),
    config_path: Path = typer.Option(
        CONFIG_PATH,
        "--config",
        "-c",
        help="Path to config.ini",
        exists=True,
        dir_okay=False,
    ),
    quiet: bool = typer.Option(False, "--quiet", "-q", help="Quiet mode: less output"),
    debugLevel: int = typer.Option(
        logging.NOTSET,
        "--debug",
        "-d",
        help="Debug level",
        min=0,
        max=5,
        clamp=True,
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Dry run: do not send mails"
    ),
):
    """send emails to a list of recipients as configured in your letter"""
    if letter_path is None:
        letter_names = list(
            filter(
                lambda letter: Path(letter).is_dir(),
                os.listdir("."),
            )
        )

        if len(letter_names) == 0:
            richError(
                f"Can't find any letters in {os.getcwd()}, please specify a letter path"
            )

        letter_name = typerSelect("Please select a letter", letter_names)
        letter_path = Path(letter_name).absolute()

    print(f"Using letter [blue]{letter_path}\n")

    if not Letter.check_letter(letter_path, verbose=not quiet):
        richError(f"Invalid letter: {letter_path}")
        return

    setup_logger(letter_path / "log.txt", debugLevel)

    if test_mode:
        print("[blue]Entering test mode...\n")
        time.sleep(1)

    if dry_run:
        print("[blue]Entering dry run mode...\n")
        time.sleep(1)

    auto_mailer_config = AutoMailer.load_mailer_config(config_path)
    auto_mailer = AutoMailer(auto_mailer_config, quiet=quiet)
    emails = Letter(
        letter_path,
        auto_mailer_config["account"]["name"],
        test_mode=test_mode,
    )
    auto_mailer.login()
    auto_mailer.send_emails(emails, test_mode=test_mode, dry=dry_run)
    auto_mailer.check_bounce_backs()
    richSuccess(
        f"{auto_mailer.success_count} / {auto_mailer.total_count} emails sent successfully"
    )


@app.command()
def new(letter_name: Optional[str] = typer.Argument(..., help="Name of letter")):
    """create a new letter from template"""
    print("Creating a new letter")

    TEMPLATE_PATH = APP_ROOT / "template_letter"

    if letter_name is None:
        letter_name = Prompt.ask("Please enter the letter name")
    try:
        shutil.copytree(TEMPLATE_PATH, Path(os.getcwd()) / letter_name)
    except FileExistsError:
        richError("directory already exists")
        return

    richSuccess(f"Letter {letter_name} created")


@app.command()
def check(
    letter_path: Path = typer.Argument(
        ...,
        help="Path to letter directory",
        exists=True,
        file_okay=False,
    ),
):
    """
    check wether a directory is a valid letter\n
    a letter folder should be structured as follows:\n
    ```\n
    letter_name\n
    ├── attachments\n
    │   ├── ...\n
    │   └── ...\n
    ├── config.yml\n
    ├── content.html\n
    └── recipients.csv\n
    ```\n
    """

    print("Checking letter")
    if Letter.check_letter(letter_path, verbose=True):
        richSuccess("Letter is valid")
    else:
        richError("Letter is invalid")


@app.command()
def config(
    new_config_path: Optional[str] = typer.Option(
        None,
        "--file",
        "-f",
        help="Path to new config file whose content will be copied to config.ini",
        exists=True,
        dir_okay=False,
    ),
    reset: bool = typer.Option(
        False, "--reset", "-r", help="Reset config.ini to default"
    ),
    list_config: Optional[bool] = typer.Option(
        False, "--list", "-l", help="list current config"
    ),
):
    """
    configure the auto mailer\n
    a valid config file should have the following structure:\n
    [smtp]\n
    host=smtps.ntu.edu.tw\n
    port=465\n
    timeout=5\n
    [pop3]\n
    host=msa.ntu.edu.tw\n
    port=995\n
    timeout=5\n
    [account]\n
    name=John Doe\n
    """

    if list_config:
        typer.echo(Path(CONFIG_PATH).read_text())
        return

    if reset:
        shutil.copy(APP_ROOT / "config-default.ini", CONFIG_PATH)
        richSuccess("Config file reset to default")
        return

    APP_DIR.mkdir(parents=True, exist_ok=True)

    if new_config_path is not None:
        if not AutoMailer.validate_config(CONFIG_PATH):
            richError("Invalid config file")
            exit(1)

        shutil.copy(new_config_path, CONFIG_PATH)
        richSuccess(f"Config file copied to {CONFIG_PATH}")
        return

    config = AutoMailer.load_mailer_config(CONFIG_PATH)

    for section, vals in config.items():
        for key, value in vals.items():
            print(f"\n{section}.{key} = {value}")
            if Confirm.ask(
                f"Do you want to modify [blue]{section}.{key}[/blue]?", default=False
            ):
                new_value = Prompt.ask(
                    f"Enter new value for [blue]{section}.{key}",
                    password=(key == "password"),
                )
                new_value = new_value.encode(ENCODING).decode("utf-8")
                config[section][key] = new_value
                richSuccess(f"{section}.{key} updated")

    richSuccess(f"Config file updated to {CONFIG_PATH}")



##########################################################################
# File:         Letter.py                                                #
# Purpose:      Automatically send batch of mails                        #
# Last changed: 2015/06/21                                               #
# Author:       zhuang-jia-xu                                            #
# Edited:                                                                #
# Copyleft:     (ɔ)NTUEE                                                 #
##########################################################################
from cerberus import Validator
from email_validator import validate_email, caching_resolver

import os
import csv
import yaml
import re
import logging
from typing import List
from pathlib import Path, PurePath
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from email.utils import formataddr, formatdate
from string import Template

from .utils import *

__all__ = ["Letter"]

letter_config_schema = {
    "subject": {"type": "string", "required": True},
    "from": {"type": "string"},
    "recipientTitle": {"type": "string"},
    "lastNameOnly": {"type": "boolean"},
    "cc": {"type": "list"},
    "bcc": {"type": "list"},
    "bccToSender": {"type": "boolean"},
}

v = Validator(letter_config_schema)

RESERVED_FIELDS = ("email", "cc", "bcc")
REQUIRED_FIELDS = ("email", "name")

class Letter:
    paths: dict = None
    config: dict = None
    csv: List = None
    email_addrs: List = None
    emails: List = None

    def __init__(self, letter_path: str, sender_name: str, *, test_mode: bool = False):
        if not self.validate_letter_dir(letter_path, verbose=True):
            richError(f"{letter_path} is not a valid letter directory")

        self.paths = self.get_paths(letter_path)

        self.config = {
            **self.__load_letter_config(),
            "sender_name": sender_name,
            "attachments": [],
        }

        if Path(self.paths["attachments"]).is_dir():
            attachments = list(
                filter(lambda a: a[0] != ".", os.listdir(self.paths["attachments"]))
            )
            if len(attachments) > 0:
                self.config["attachments"] = [
                    Path(self.paths["attachments"]) / a for a in attachments
                ]

        self.csv = self.__load_recipients()
        self.email_addrs = [row["email"] for row in self.csv]
        self.emails = self.__generate_emails(test_mode=test_mode)

    def set_from_addr(self, from_addr: str):
        for email in self.emails:
            email["From"] = formataddr((self.config["from"], from_addr))

            if "bcc" in self.config:
                if "bccToSender" in self.config and self.config["bccToSender"]:
                    bccs = [*self.config["bcc"], from_addr]
                    email["Bcc"] = ",".join(bccs)
                else:
                    email["Bcc"] = ",".join(self.config["bcc"])
            elif "bccToSender" in self.config and self.config["bccToSender"]:
                email["Bcc"] = from_addr

    def __load_letter_config(self):
        path = self.paths["config"]
        if not Path(path).is_file():
            logging.error(f"{path} is not a file")
            richError(
                f"failed to load letter config at {path}, please enter a valid letter path"
            )

        letter_config = self.load_file(path)

        if not self.validate_letter_config(letter_config, verbose=True):
            logging.error(f"{path} is not a valid letter config")
            richError(
                f"failed to load letter config at {path}, please enter a valid letter path"
            )

        if "cc" in letter_config:
            letter_config["cc"] = letter_config["cc"] = complete_school_email(
                letter_config["cc"]
            )
        if "bcc" in letter_config:
            letter_config["bcc"] = complete_school_email(letter_config["bcc"])

        return letter_config

    def __load_recipients(self):
        """load recipients from csv file"""
        path = self.paths["recipients"]
        recipients = self.load_file(path)

        is_valid = self.validate_recipients(recipients, verbose=True)

        if not is_valid:
            logging.error(f"letter csv: {Path(self.paths['recipients']).read_text()}")
            richError(f"failed to load recipients from {self.paths['recipients']}")
            return

        return recipients

    def __generate_emails(self, *, test_mode: bool = False):
        """generate emails from csv file"""
        emails = []

        email_template = Path(self.paths["content"]).read_text(encoding=ENCODING)

        if not self.validate_email_content(email_template, self.csv[0].keys(), verbose=True):
            return
        
        email_template = Template(email_template)

        # create attachments
        mime_attachments = []
        for attachment in self.config["attachments"]:
            with open(attachment, "rb") as f:
                mime_attachment = MIMEApplication(
                    f.read(), Name=os.path.basename(attachment)
                )

            mime_attachment[
                "Content-Disposition"
            ] = f"attachment; filename={os.path.basename(attachment)}"

            mime_attachments.append(mime_attachment)

        for recipient in self.csv:
            email = self.__generate_email(recipient, email_template, mime_attachments)
            emails.append(email)
            if test_mode:
                break
        return emails

    def __generate_email(
        self,
        recipient: dict,
        email_template: Template,
        mime_attachments: List[MIMEApplication],
    ):
        """generate email from recipient"""

        email = MIMEMultipart()
        email["Date"] = formatdate(localtime=True)

        email["Subject"] = self.config["subject"]

        email["To"] = recipient["email"]

        cc_list = []
        if "cc" in self.config:
            cc_list += self.config["cc"]
        if "cc" in recipient:
            cc_list += recipient["cc"].split(" ")
        if len(cc_list) > 0: 
            email["Cc"] = ",".join(cc_list)

        bcc_list = []
        if "bcc" in self.config:
            bcc_list += self.config["bcc"]
        if "bcc" in recipient:
            bcc_list += recipient["bcc"].split(" ")
        if len(bcc_list) > 0:
            email["Bcc"] = ",".join(bcc_list)

        if "recipientTitle" in self.config:
            if "lastNameOnly" in self.config and self.config["lastNameOnly"]:
                # only supports chinese names
                recipient["name"] = recipient["name"][0]
            recipient["name"] = recipient["name"] + self.config["recipientTitle"]

        email.attach(
            MIMEText(
                email_template.substitute(
                    {**recipient, "sender": self.config["sender_name"]}
                ),
                "html",
            )
        )

        for mime_attachment in mime_attachments:
            email.attach(mime_attachment)

        return email

    def __iter__(self):
        return iter(self.emails)

    def __len__(self):
        return len(self.emails)

    @classmethod
    def load_file(cls, file_path: str):
        """load letter from config"""
        file_name = PurePath(file_path).name
        if file_name == "content.html":
            return Path(file_path).read_text(encoding=ENCODING)

        elif file_name == "config.yml" or file_name == "config.yaml":
            with open(file_path, encoding=ENCODING) as f:
                letter_config = yaml.load(f, Loader=yaml.FullLoader)

            return letter_config

        elif file_name == "recipients.csv":
            with open(file_path, encoding=ENCODING) as f:
                reader = csv.DictReader(f)
                recipients = [row for row in reader]

                stripped_recipients = []
                for row in recipients:
                    temp_row = {}
                    for key, value in row.items():
                        temp_row[key.strip()] = value.strip()

                    if "email" in temp_row:
                        temp_row["email"] = complete_school_email(temp_row["email"].lower())
                    if "cc" in temp_row:
                        temp_row["cc"] = complete_school_email(temp_row["cc"].lower())
                    if "bcc" in temp_row:
                        temp_row["bcc"] = complete_school_email(temp_row["bcc"].lower())
                    
                    stripped_recipients.append(temp_row)

            return stripped_recipients

        else:
            return None

    @classmethod
    def check_letter(cls, letter_path: str, verbose=False) -> bool:
        """check letter"""
        is_valid = True

        is_valid &= cls.validate_letter_dir(letter_path, verbose=verbose)
        if not is_valid:
            return False

        paths = cls.get_paths(letter_path)

        config_file = cls.load_file(paths["config"])
        content_file = cls.load_file(paths["content"])
        recipients_file = cls.load_file(paths["recipients"])

        is_valid &= cls.validate_letter_config(config_file, verbose=verbose)

        is_valid &= cls.validate_recipients(recipients_file, verbose=verbose)
        if not is_valid:
            return False

        is_valid &= cls.validate_email_content(
            content_file, recipients_file[0], verbose=verbose
        )

        return is_valid

    @classmethod
    def validate_letter_dir(cls, letter_path: str, verbose=False) -> bool:
        paths = cls.get_paths(letter_path)

        is_valid = True
        for key, path in paths.items():
            if not path.exists():
                if verbose:
                    logging.error(f"letter: {path} not found")
                    richError(f"letter {key} not found at {path}", terminate=False)
                    is_valid = False
                else:
                    return False
            if key == "attachments":
                if not path.is_dir():
                    if verbose:
                        richError(f"{path} should be a dir", terminate=False)
                        is_valid = False
                    else:
                        return False
            else:
                if not path.is_file():
                    if verbose:
                        richError(f"{path} should be a file", terminate=False)
                        is_valid = False
                    else:
                        return False

        return is_valid

    @classmethod
    def validate_letter_config(cls, letter_config: dict, verbose=False) -> bool:
        is_valid = v.validate(letter_config)
        if verbose:
            parse_validation_error(v._errors)
        return is_valid

    @classmethod
    def validate_recipients(cls, stripped_recipients: list, verbose=False) -> bool:
        """validate recipients"""

        is_valid = True

        for required in REQUIRED_FIELDS:
            if required not in stripped_recipients[0]:
                if verbose:
                    logging.error(
                        f"{required} is a required field in the csv file, but not found"
                    )
                    richError(
                        f"{required} is a required field in the csv file, but not found",
                        terminate=False,
                    )
                    is_valid = False
                else:
                    return False
                    
        for row in stripped_recipients:
            for i, (key, value) in enumerate(row.items()):
                if key is None:
                    if verbose:
                        logging.error(f"too many fields at row {i} in recipients.csv")
                        richError(
                            f"too many fields at row {i} in recipients.csv",
                            terminate=False,
                        )
                        is_valid = False
                        continue
                    else:
                        return False

                if key in REQUIRED_FIELDS and value == "":
                    if verbose:
                        logging.error(
                            f"{key} cannot be empty, recipients.csv has no {key} at row {i}"
                        )
                        richError(
                            f"{key} cannot be empty, recipients.csv has no {key} at row {i}",
                            terminate=False,
                        )
                        is_valid = False
                    else:
                        return False

        resolver = caching_resolver()

        for i, row in enumerate(stripped_recipients):
            row["email"] = complete_school_email(row["email"].lower())
            try:
                validate_email(row["email"], dns_resolver=resolver)
            except:
                if verbose:
                    logging.error(
                        f"recipients.csv has invalid email: {row['email']}, at row {i}"
                    )
                    richError(
                        f"invalid email {row['email']} detected at row {i} in recipients.csv",
                        terminate=False,
                    )
                    is_valid = False
                else:
                    return False

        return is_valid

    @classmethod
    def validate_email_content(
        cls, email_template: str, csv_indexes: list, verbose=False
    ) -> bool:
        """validate email content"""
        template_fields = re.findall(r"\$([_a-z][_a-z0-9]*)", email_template, re.M)
        is_valid = True

        for reserved in RESERVED_FIELDS:
            if reserved in template_fields:
                if verbose:
                    logging.error(
                        f"{reserved} is a reserved field, it cannot be used in email template"
                    )
                    richError(
                        f"{reserved} is a reserved field, it cannot be used in email template",
                        terminate=False,
                    )
                    is_valid = False
                else:
                    return False

        csv_indexes = [f.strip() for f in template_fields] + ["sender"]
        for field in template_fields:
            if field not in csv_indexes:
                if verbose:
                    logging.error(f"letter template has field '{field}', but is not found in csv")
                    richError(
                        f"letter template has field '{field}', but is not found in csv",
                        terminate=False,
                    )
                    is_valid = False
                else:
                    return False

        return is_valid

    @classmethod
    def get_paths(self, letter_path: str) -> list:
        """get paths to different part of letters"""
        letter_root_path = Path(letter_path)

        paths = {
            "content": letter_root_path / "content.html",
            "config": letter_root_path / "config.yml",
            "attachments": letter_root_path / "attachments",
            "recipients": letter_root_path / "recipients.csv",
        }

        return paths



##########################################################################
# File:         utils.py                                                 #
# Purpose:      Automatically send batch of mails                        #
# Last changed: 2015/06/21                                               #
# Author:       zhuang-jia-xu                                            #
# Edited:                                                                #
# Copyleft:     (ɔ)NTUEE                                                 #
##########################################################################
import typer
from rich import print
from rich.progress import Progress, TextColumn
from rich.prompt import Confirm
from configparser import ConfigParser
from cerberus.errors import ValidationError, ErrorList

import logging
import time
from pathlib import Path

from .globals import *


def richError(
    *objects: any, end: str = "\n", prefix: str = "Error: ", terminate: bool = True
) -> None:
    print(f"[red]{prefix}", end="")
    for o in objects:
        print(f"[red]{o}", end="")
    print("", end=end)
    if terminate:
        exit(1)


def richWarning(text: str, end: str = "\n", prefix: bool = True) -> None:
    if prefix:
        print(f"[yellow]Warning: [/yellow]", end="")
    print(f"[yellow]{text}[/yellow]", end=end)


def richSuccess(text: str, end: str = "\n") -> None:
    print(f"[green]{text}[/green]", end=end)


def parse_validation_error(errors: ErrorList) -> None:
    def _parse(error: ValidationError, indent: int = 1) -> None:
        if type(error) == ErrorList:
            for e in error:
                _parse(e, indent)
            return

        tabs = " " * 4 * indent

        if error.rule == "schema":
            richError(
                f"{tabs}{'/'.join(error.document_path)}: missing entry",
                terminate=False,
                prefix=False,
            )
        elif error.rule == "type":
            richError(
                f"{tabs}{'/'.join(error.document_path)}: should be {error.constraint}",
                terminate=False,
                prefix=False,
            )
        elif error.rule == "coerce":
            richError(
                f"{tabs}{'/'.join(error.document_path)}: ",
                *error.info,
                terminate=False,
                prefix=False,
            )
        elif error.rule == "required":
            richError(
                f"{tabs}{'/'.join(error.document_path)}: required",
                terminate=False,
                prefix=False,
            )
        else:
            richError(
                f"{tabs}{'/'.join(error.document_path)}: ",
                *error.info,
                terminate=False,
                prefix=False,
            )

        if type(error.info) == tuple and len(error.info) > 0:
            for e in error.info:
                if type(e) != str:
                    _parse(e, indent + 1)

    _parse(errors)


def complete_school_email(email_addr) -> list:
    """
    Autocomplete school email addresss
    """
    if type(email_addr) == list:
        return [complete_school_email(addr) for addr in email_addr]

    if "@" not in email_addr:
        return f"{email_addr}@ntu.edu.tw"
    else:
        return email_addr


def typerSelect(message: str, options: list) -> str:
    def process_options(n):
        n = int(n)
        if n < 0 or n >= len(options):
            raise typer.BadParameter("invalid letter index")
        return options[n]

    prompt_message = (
        "\n"
        + "\n".join(
            [
                f"{'> ' if i == 0 else '  '}[{i}] {option}"
                for i, option in enumerate(options)
            ]
        )
        + "\n\n"
        + message
    )

    result = typer.prompt(
        prompt_message,
        value_proc=process_options,
        default=0,
    )
    print()
    return result


def countdownConfirm(message: str, default=True, countdown: int = 0) -> bool:
    prompt_message = f"\n{message}"
    if countdown == 0:
        result = Confirm.ask(prompt_message, default=default)
    else:
        with Progress(
            TextColumn("[bold blue]{task.description}"), transient=True
        ) as progress:
            countdown_task = progress.add_task("", total=countdown)
            for i in range(countdown + 1):
                progress.update(
                    countdown_task,
                    total=i + 1,
                    description=f' {countdown - i}{" ." * (countdown - i)}',
                )
                time.sleep(1)
        result = Confirm.ask(prompt_message, default=default)
    print()
    return result


DEBUG_LEVELS = [
    logging.NOTSET,
    logging.DEBUG,
    logging.INFO,
    logging.WARNING,
    logging.ERROR,
    logging.CRITICAL,
]


def setup_logger(file_path, level):
    # create log file if it doesn't exist
    if not Path(file_path).is_file():
        with open(file_path, "w") as f:
            pass

    logging.basicConfig(
        level=DEBUG_LEVELS[level],
        filename=file_path,
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%Y/%m/%d %I:%M:%S %p",
    )


if __name__ == "__main__":
    richSuccess("Success")
    richWarning("Warning")
    richError("Error")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ntuee_mailer']

package_data = \
{'': ['*'],
 'ntuee_mailer': ['template_letter/*', 'template_letter/attachments/*']}

install_requires = \
['Cerberus==1.3.4',
 'PyYAML==6.0',
 'click>=8.1.3,<9.0.0',
 'email-validator==1.2.1',
 'rich==12.5.1',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['ntuee-mailer = ntuee_mailer.main:app']}

setup_kwargs = {
    'name': 'ntuee-mailer',
    'version': '2.0.12',
    'description': 'an auto mailer to send emails in batch for you',
    'long_description': '# `ntuee-mailer`\n\nThis is a simple mailer for NTU students to send letters in batches.\n\n**Installation**\n\n```bash\n$ pip install ntuee-mailer\n```\n\n**Usage**:\n\n```bash\n$ ntuee-mailer [OPTIONS] COMMAND [ARGS]...\n```\n\n**Options**:\n\n- `--install-completion`: Install completion for the current shell.\n- `--show-completion`: Show completion for the current shell, to copy it or customize the installation.\n- `--help`: Show this message and exit.\n\n**Commands**:\n\n- `check`: check wether a directory is a valid letter a...\n- `config`: configure the auto mailer a valid config file...\n- `new`: create a new letter from template\n- `send`: send emails to a list of recipients as...\n\n## `ntuee-mailer check`\n\ncheck wether a directory is a valid letter\n\na letter folder should be structured as follows:\n\n```\nletter_name\n├── attachments\n│ ├── ...\n│ └── ...\n├── config.yml\n├── content.html\n└── recipients.csv\n```\n\n**Usage**:\n\n```console\n$ ntuee-mailer check [OPTIONS] LETTER_PATH\n```\n\n**Arguments**:\n\n- `LETTER_PATH`: Path to letter directory [required]\n\n**Options**:\n\n- `--help`: Show this message and exit.\n\n## `ntuee-mailer config`\n\nconfigure the auto mailer\n\na valid config file should have the following structure:\n\n[smtp]\n\nhost=smtps.ntu.edu.tw\n\nport=465\n\ntimeout=5\n\n[pop3]\n\nhost=msa.ntu.edu.tw\n\nport=995\n\ntimeout=5\n\n[account]\n\nname=John Doe\n\n**Usage**:\n\n```console\n$ ntuee-mailer config [OPTIONS]\n```\n\n**Options**:\n\n- `-f, --file TEXT`: Path to new config file whose content will be copied to config.ini\n- `-r, --reset`: Reset config.ini to default [default: False]\n- `-l, --list`: list current config [default: False]\n- `--help`: Show this message and exit.\n\n## `ntuee-mailer new`\n\ncreate a new letter from template\n\n**Usage**:\n\n```console\n$ ntuee-mailer new [OPTIONS] LETTER_NAME\n```\n\n**Arguments**:\n\n- `LETTER_NAME`: Name of letter [required]\n\n**Options**:\n\n- `--help`: Show this message and exit.\n\n## `ntuee-mailer send`\n\nsend emails to a list of recipients as configured in your letter\n\n**Usage**:\n\n```console\n$ ntuee-mailer send [OPTIONS] [LETTER_PATH]\n```\n\n**Arguments**:\n\n- `[LETTER_PATH]`: Path to letter\n\n**Options**:\n\n- `-t, --test`: Test mode: send mail to yourself [default: False]\n- `-c, --config FILE`: Path to config.ini [default: /home/madmax/.config/ntuee-mailer/config.ini]\n- `-q, --quiet`: Quiet mode: less output [default: False]\n- `-d, --debug INTEGER RANGE`: Debug level [default: 0]\n- `--help`: Show this message and exit.\n\n## `ntuee-mailer test`\n\n**Usage**:\n\n```console\n$ ntuee-mailer test [OPTIONS]\n```\n\n**Options**:\n\n- `--help`: Show this message and exit.\n\n## mail format\n\na letter folder should be structured as follows:\n```\nletter_name\n├── attachments\n│ ├── ...\n│ └── ...\n├── config.yml\n├── content.html\n└── recipients.csv\n```\n\n### content.html\nThe content of the email. `$<pattern>` would be replaced by the corresponding field defined in `recipients.csv`\n\n### recipients.csv\nStores the data related to recipients. The value of "name" field is will be used to replace `$name` in `content.html`, whose behavior can be modified in `config.yml`. The "email" field stores the recipients email. The emails will be CCed and BCCed to the emails in "cc" and "bcc" field. One recipients may have several CC and BCCs, emails should be separated with spaces. "email", "cc" and "bcc" are reserved fields, they cannot be used in html pattern, any additional field will be replaced in the html. "name" and "email" fields are required\n\n### config.yml\nConfiguration of each email. "subjects" defines subject, "from" defines the name recipients see in their email client. "recipientTitle" and "lastNameOnly" modifies the behavior of `$name` in `content.html`.\n\n### attachments\nThe attachment directory. Any file placed in this folder will be attached to the email. Any file with name started with \'.\' will be ignored, i.e. .git, .DS_STORE.\n',
    'author': 'madmaxieee',
    'author_email': '76544194+madmaxieee@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/madmaxieee/ntuee-mailer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
